<?php

namespace Drupal\entity_sync_ui\Controller;

use Drupal\entity_sync\Config\ManagerInterface as ConfigManagerInterface;
use Drupal\entity_sync\Entity\SyncInterface;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;

use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides pages with synchronization data related to individual entities.
 */
class Entity extends ControllerBase {

  /**
   * The Entity Synchronization configuration manager.
   *
   * @var \Drupal\entity_sync\Config\ManagerInterface
   */
  protected $configManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new Entity object.
   *
   * @param \Drupal\entity_sync\Config\ManagerInterface $config_manager
   *   The Entity Sync configuration manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    ConfigManagerInterface $config_manager,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->configManager = $config_manager;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_sync.config_manager'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Returns the page that renders basic synchronization info for the entity.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   A RouteMatch object.
   *
   * @return array
   *   The render array for the page.
   */
  public function infoPage(RouteMatchInterface $route_match) {
    $entity = $this->getEntityFromRouteMatch($route_match);
    if (!$entity) {
      throw new \RuntimeException(
        'The sync information page can only be rendered on entity pages.'
      );
    }

    // Get all synchronizations for the entity type.
    $syncs = $this->configManager->getSyncs([
      'local_entity' => ['type_id' => $entity->getEntityTypeId()],
    ]);
    if (!$syncs) {
      return [
        '#markup' => $this->t(
          'There are no synchronizations for this entity type.'
        ),
      ];
    }

    $build = [];

    foreach ($syncs as $sync) {
      $build[$sync->id()] = $this->buildSynchronizationDetails($entity, $sync);
    }

    return $build;
  }

  /**
   * Retrieves the entity from route match.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The entity object, or NULL if not found in the given route.
   */
  protected function getEntityFromRouteMatch(RouteMatchInterface $route_match) {
    return $route_match->getParameter(
      $route_match->getRouteObject()->getOption('_entity_sync_entity_type_id')
    );
  }

  /**
   * Builds the render array for the given entity/synchronization pair.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The synchronization.
   */
  protected function buildSynchronizationDetails(
    EntityInterface $entity,
    SyncInterface $sync
  ) {
    $element = [
      '#type' => 'details',
      '#title' => $sync->label(),
      '#open' => TRUE,
    ];

    $element['description'] = [
      '#markup' => '<p>' . $sync->get('description') . '</p>',
    ];

    // Status.
    $status = $this->t('Disabled');
    if ($sync->get('status')) {
      $status = $this->t('Enabled');
    }
    $element['status'] = [
      '#markup' => $this->t(
        '<strong>Status:</strong> @status',
        ['@status' => $status]
      ),
    ];

    // Linked entity ID.
    $field_name = $sync->getLocalEntitySettings()['remote_id_field'] ?? NULL;
    $id_field = $entity->get($field_name);
    if ($id_field->isEmpty()) {
      $element['remote_id'] = [
        '#markup' => $this->t(
          '<br><strong>Linked entity ID:</strong> Not linked'
        ),
      ];
    }
    else {
      $element['remote_id'] = [
        '#markup' => $this->t(
          '<br><strong>Linked entity ID:</strong> @remote_id',
          ['@remote_id' => $id_field->value]
        ),
      ];
    }

    return $element;
  }

}
