<?php

namespace Drupal\entity_sync\Export;

use Drupal\entity_sync\Client\ClientFactory;
use Drupal\entity_sync\Config\ManagerInterface as ConfigManagerInterface;
use Drupal\entity_sync\Entity\OperationTypeInterface;
use Drupal\entity_sync\Entity\SyncInterface;
use Drupal\entity_sync\Event\QueueOperationEvent;
use Drupal\entity_sync\Export\Event\Events;
use Drupal\entity_sync\Export\Event\LocalEntityMappingEvent;
use Drupal\entity_sync\EntityManagerBase;
use Drupal\entity_sync\MachineName\Field\Operation as OperationField;
use Drupal\entity_sync\StateManagerInterface;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Queue\QueueFactory;

use Psr\Log\LoggerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * The default export manager.
 */
class EntityManager extends EntityManagerBase implements EntityManagerInterface {

  /**
   * The client factory.
   *
   * @var \Drupal\entity_sync\Client\ClientFactory
   */
  protected $clientFactory;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The Entity Sync configuration manager.
   *
   * @var \Drupal\entity_sync\Config\ManagerInterface
   */
  protected $configManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The Entity Sync export field manager.
   *
   * @var \Drupal\entity_sync\Export\FieldManagerInterface
   */
  protected $fieldManager;

  /**
   * The Entity Sync state manager.
   *
   * @var \Drupal\entity_sync\StateManagerInterface
   */
  protected $stateManager;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Constructs a new EntityManager instance.
   *
   * @param \Drupal\entity_sync\Client\ClientFactory $client_factory
   *   The client factory.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\entity_sync\Config\ManagerInterface $config_manager
   *   The Entity Sync configuration manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\entity_sync\Export\FieldManagerInterface $field_manager
   *   The Entity Sync export field manager.
   * @param \Drupal\entity_sync\StateManagerInterface $state_manager
   *   The Entity Sync export field manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   The queue factory.
   */
  public function __construct(
    ClientFactory $client_factory,
    ConfigFactoryInterface $config_factory,
    ConfigManagerInterface $config_manager,
    EntityTypeManagerInterface $entity_type_manager,
    EventDispatcherInterface $event_dispatcher,
    FieldManagerInterface $field_manager,
    StateManagerInterface $state_manager,
    LoggerInterface $logger,
    QueueFactory $queue_factory
  ) {
    $this->clientFactory = $client_factory;
    $this->configFactory = $config_factory;
    $this->configManager = $config_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->eventDispatcher = $event_dispatcher;
    $this->fieldManager = $field_manager;
    $this->stateManager = $state_manager;
    $this->logger = $logger;
    $this->queueFactory = $queue_factory;
  }

  /**
   * {@inheritdoc}
   */
  public function exportLocalListByEntities(
    $sync_id,
    array $local_entities,
    array $options = []
  ) {
    // Load the sync.
    // phpcs:disable
    // @I Validate the sync/operation
    //    type     : bug
    //    priority : normal
    //    labels   : operation, sync, validation
    //    notes    : Review whether the validation should happen upon runtime
    //               i.e. here, or when the configuration is created/imported.
    // @I Validate that the provider supports the `export_list` operation
    //    type     : bug
    //    priority : normal
    //    labels   : operation, sync, validation
    //    notes    : Review whether the validation should happen upon runtime
    //               i.e. here, or when the configuration is created/imported.
    // @I Validate that the given local entities are of the correct type
    //    type     : bug
    //    priority : normal
    //    labels   : export, operation, sync, validation
    // phpcs:enable
    $sync = $this->entityTypeManager->getStorage('entity_sync')->load($sync_id);
    if (!$sync) {
      $sync = $this->entityTypeManager
        ->getStorage('entity_sync_operation_type')
        ->loadWithPluginInstantiated($sync_id);
    }

    // Make sure the operation is enabled and supported by the provider.
    if (!$this->operationSupported($sync, 'export_list')) {
      $this->logger->error(
        sprintf(
          'The synchronization with ID "%s" and/or its provider do not support the `export_list` operation.',
          $sync_id
        )
      );
      return;
    }

    $context = $options['context'] ?? [];

    // Notify subscribers that the operation is about to be initiated.
    // phpcs:disable
    // @I Write tests for operation cancellations
    //    type     : task
    //    priority : high
    //    labels   : export, testing
    // phpcs:enable
    $cancel = $this->preInitiate(
      Events::LOCAL_LIST_PRE_INITIATE,
      'export_list',
      ['local_entities' => $local_entities] + $context,
      $sync
    );
    if ($cancel) {
      return;
    }

    // Run the operation.
    // We do this in a `try/finally` structure so that we can still dispatch the
    // post-terminate event. Subscribers may still need to run whether the
    // operation was successfull or not. For example, even if a managed
    // operation failed we unlock it so that the next one is allowed to run.
    // At the end, the error/exception is still thrown so that the caller can
    // handle it as required.
    try {
      $this->doExportLocalList($sync, $local_entities, $options, $context);
    }
    finally {
      // Notify subscribers that the operation has terminated.
      $this->postTerminate(
        Events::LOCAL_LIST_POST_TERMINATE,
        'export_list',
        ['local_entities' => $local_entities] + $context,
        $sync
      );
    }
  }

  /**
   * {@inheritDoc}
   */
  public function exportLocalEntity(
    $sync_id,
    ContentEntityInterface $local_entity,
    array $options = []
  ) {
    // Load the sync.
    // phpcs:disable
    // @I Validate the sync/operation
    //    type     : bug
    //    priority : normal
    //    labels   : operation, sync, validation
    //    notes    : Review whether the validation should happen upon runtime
    //               i.e. here, or when the configuration is created/imported.
    // @I Validate that the provider supports the `export_entity` operation
    //    type     : bug
    //    priority : normal
    //    labels   : operation, sync, validation
    //    notes    : Review whether the validation should happen upon runtime
    //               i.e. here, or when the configuration is created/imported.
    // @I Validate that the given local entity is of the correct type
    //    type     : bug
    //    priority : normal
    //    labels   : export, operation, sync, validation
    // phpcs:enable
    $sync = $this->entityTypeManager->getStorage('entity_sync')->load($sync_id);
    if (!$sync) {
      $sync = $this->entityTypeManager
        ->getStorage('entity_sync_operation_type')
        ->loadWithPluginInstantiated($sync_id);
    }

    // Make sure the operation is enabled and supported by the provider.
    if (!$this->operationSupported($sync, 'export_entity')) {
      $this->logger->error(
        sprintf(
          'The synchronization with ID "%s" and/or its provider do not support the `export_entity` operation.',
          $sync_id
        )
      );
      return;
    }

    $this->wrapDoExportLocalEntity($sync, $local_entity, $options);
  }

  /**
   * {@inheritdoc}
   *
   * phpcs:disable
   * @I Create a queue manager or similar service for facilitating queueing
   *    type     : feature
   *    priority : normal
   *    labels   : queue
   * phpcs:enable
   */
  public function queueExportLocalEntityAllSyncs(
    ContentEntityInterface $entity,
    array $context = []
  ) {
    // Get all syncs that define exports for the given entity's type.
    // phpcs:disable
    // @I Support caching syncs per entity type
    //    type     : improvement
    //    priority : normal
    //    labels   : export, performance
    //
    // @I Support filtering by bundle
    //    type     : bug
    //    priority : high
    //    labels   : config, export
    //    notes    : The entity bundle is optional in the synchronization
    //               configuration; however, if it is defined, it should be
    //               interpreted as if the export should only proceed if the
    //               bundle of the entity being exported is of the configured
    //               type.
    // phpcs:enable
    $syncs = $this->configManager->getSyncs([
      'status' => TRUE,
      'local_entity' => [
        'type_id' => $entity->getEntityTypeId(),
      ],
      'operation' => [
        'id' => 'export_entity',
        'status' => TRUE,
      ],
    ]);
    if (!$syncs) {
      return;
    }

    // Sort the synchronizations so that they are queued in the requested
    // way. If no queue weights are defined, they will be loaded and queued in
    // alphabetical order.
    $syncs = $this->sortQueueSynchronizations($syncs);

    // The most common case is to call this method in a insert/update hook. At
    // that point the entity has already been saved and the entity's `isNew()`
    // method will return `FALSE`. We therefore expect to be given the original
    // entity and use that to detect whether we are responding to a local entity
    // creation or update i.e. whether we should be detecting field changes or
    // not.
    // phpcs:disable
    // @I Validate the original entity passed through context
    //    type     : bug
    //    priority : low
    //    labels   : export, validation
    // @I Enable/disable field comparison via options
    //    type     : improvement
    //    priority : normal
    //    labels   : export
    // phpcs:enable
    $is_update = isset($context['original_entity']) ? TRUE : FALSE;
    $changed_field_names = [];

    // Detect whether we have changed fields. If the operation is managed by
    // Entity Sync for a given synchronization and there are not changed fields,
    // we will be skipping the export.
    //
    // phpcs:disable
    // @I Support forcing an export even if unchanged via config export mode
    //    type     : improvement
    //    priority : low
    //    labels   : config, export
    // phpcs:enable
    if ($is_update) {
      $changed_field_names = $this->fieldManager->getChangedNames(
        $entity,
        $context['original_entity']
      );
    }

    foreach ($syncs as $sync) {
      $operations_settings = $sync->getOperationsSettings();

      // Do not proceed if queueing the exports is explicitly disabled.
      if (($operations_settings['export_entity']['queue'] ?? TRUE) === FALSE) {
        continue;
      }

      $is_managed = $this->stateManager->isManaged(
        $sync->id(),
        'export_entity'
      );

      // Do not proceed with queueing the export if we don't have any changed
      // fields that are also set to be exported by the current synchronization.
      if ($is_managed && $is_update) {
        if (!$changed_field_names) {
          continue;
        }

        $exportable_field_names = $this->fieldManager
          ->getExportableChangedNames(
            $entity,
            $context['original_entity'],
            $sync->getFieldMapping(),
            NULL,
            $changed_field_names
          );
        if (!$exportable_field_names) {
          continue;
        }
      }

      // If we are coming from an entity update, we may be here because an
      // entity import triggered an entity update hook. In that case, we
      // don't want to export the entity as that can cause a loop i.e. export
      // the entity to the remote which may cause the entity to be made
      // available in the recently modified list of entities, causing an import
      // again which in turn will trigger a `hook_entity_update` etc.
      //
      // If we are here as a result of a change that happened within Drupal e.g.
      // an entity edited via the UI, the remote changed field would not be
      // changed. While, if we are here as a result of an import, the original
      // remote changed field would be a timestamp earlier than the new remote
      // changed field.
      //
      // We therefore check for that, and if the original remote changed field
      // is earlier than the updated remote changed field then we do not export
      // the entity.
      //
      // This will also prevent triggering another export as a result of saving
      // the entity in order to update the remote changed field when getting
      // the response from the create/update export operation.
      //
      // It does not cover though the case that we import changes that have
      // already been imported where the remote changed field will be the same
      // before and after the import. That still remains to be resolved and it
      // relies on finding a way to let the method that saves the entity provide
      // custom data to `hook_entity_insert`.
      // See https://www.drupal.org/project/drupal/issues/3158180
      //
      // Note that we may not always have a remote changed field; that can be,
      // for example, when we only export an entity type while imports are
      // disabled. We therefore skip the check if imports are disabled.
      //
      // phpcs:disable
      // @I Test skipping exports that are triggered by imports
      //    type     : task
      //    priority : high
      //    labels   : export, testing
      // @I Review case of imports+exports without a remote changed field
      //    type     : bug
      //    priority : high
      //    labels   : export
      // phpcs:enable
      $import_status = FALSE;
      if ($operations_settings['import_list']['status'] ?? NULL) {
        $import_status = TRUE;
      }
      elseif ($operations_settings['import_entity']['status'] ?? NULL) {
        $import_status = TRUE;
      }
      if ($import_status && $is_managed && $is_update) {
        $field_name = $sync->getLocalEntitySettings()['remote_changed_field'] ?? NULL;

        $original_changed = $context['original_entity']
          ->get($field_name)
          ->value;
        if ($original_changed < $entity->get($field_name)->value) {
          continue;
        }
      }

      // If we are coming here from an entity creation, we want to prevent
      // exporting entities that were created as a result of an import. In those
      // cases, the remote changed field of the new local entity will not be
      // empty, while if the entity was created within the application it will
      // always be empty.
      // See entity_sync_entity_bundle_field_info_alter().
      //
      // As above, we skip the checks if imports are disabled.
      //
      // phpcs:disable
      // @I Allow queueing exports of imported entities for different syncs
      //    type     : bug
      //    priority : high
      //    labels   : export
      //    notes    : There are legitimate cases where we still want to allow
      //               exports of new imported entities e.g. import from one
      //               remote resource and send to another.
      // phpcs:enable
      elseif ($import_status && $is_managed && !$is_update) {
        $remote_changed_field = $entity->get(
          $sync->getLocalEntitySettings()['remote_changed_field'] ?? NULL
        );

        if (!$remote_changed_field->isEmpty()) {
          continue;
        }
      }

      // Lastly, give the opportunity to event subscribers to cancel queueing
      // the item. Usually, we cancel items that shouldn't be exported at the
      // time that the operation is run via a pre-initiate event subscriber.
      // However, there are cases where we want to prevent queueing exports
      // in the first place, such as  when entities are created or updated in
      // large numbers but not all of them should be exported. Cancelling
      // before queueing helps to avoid creating too many unnecessary queue
      // items and operation entities.
      if ($this->shouldCancelQueueing($sync, $entity, $context)) {
        continue;
      }

      // Queue the export.
      // phpcs:disable
      // @I Support exporting the diff between specific revisions
      //    type     : improvement
      //    priority : normal
      //    labels   : export
      //    notes    : If the entity being exported is revisioned, we can store
      //               in the queue item the specific revisions and export the
      //               fields changed between those revisions.
      // phpcs:enable
      $this->createQueueItem($sync, $entity, 'entity_sync_export_local_entity');
    }
  }

  /**
   * Runs the actual local entity export operation.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param \Drupal\Core\Entity\ContentEntityInterface[] $local_entities
   *   The local entities.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   import. For supported options see
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalListByEntities().
   * @param array $context
   *   An associative array containing of context related to the circumstances
   *   of the operation. See
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalListByEntities().
   */
  protected function doExportLocalList(
    SyncInterface $sync,
    array $local_entities,
    array $options,
    array $context
  ) {
    // Initiate the operation.
    $this->initiate(
      Events::LOCAL_LIST_INITIATE,
      'export_list',
      ['local_entities' => $local_entities] + $context,
      $sync
    );

    // Go through each local entity and create or update the corresponding
    // remote entity.
    foreach ($local_entities as $local_entity) {
      $this->tryWrapDoExportLocalEntity(
        $local_entity,
        $sync,
        ['parent_operation' => 'export_local_list'] + $options,
        'export_list'
      );
    }

    // Terminate the operation.
    $this->terminate(
      Events::LOCAL_LIST_TERMINATE,
      'export_list',
      ['local_entities' => $local_entities] + $context,
      $sync
    );
  }

  /**
   * Runs the actual local entity export operation with its pre/post events.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   import. For supported options see the operations that we may be running:
   *   - \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalListByEntities()
   *   - \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity()
   */
  protected function wrapDoExportLocalEntity(
    SyncInterface $sync,
    ContentEntityInterface $local_entity,
    array $options
  ) {
    $context = $options['context'] ?? [];
    $context['local_entity'] = $local_entity;

    // Notify subscribers that the operation is about to be initiated.
    // phpcs:disable
    // @I Write tests for operation cancellations
    //    type     : task
    //    priority : high
    //    labels   : import, testing
    // phpcs:enable
    $cancel = $this->preInitiate(
      Events::LOCAL_ENTITY_PRE_INITIATE,
      'export_entity',
      $context,
      $sync
    );
    if ($cancel) {
      return;
    }

    // Run the operation.
    // We do this in a `try/finally` structure so that we can still dispatch the
    // post-terminate event. Subscribers may still need to run whether the
    // operation was successfull or not. For example, even if a managed
    // operation failed we unlock it so that the next one is allowed to run.
    // At the end, the error/exception is still thrown so that the caller can
    // handle it as required.
    try {
      $this->doExportLocalEntity($sync, $local_entity, $options, $context);
    }
    finally {
      // Notify subscribers that the operation has terminated.
      $this->postTerminate(
        Events::LOCAL_ENTITY_POST_TERMINATE,
        'export_entity',
        $context,
        $sync
      );
    }
  }

  /**
   * Runs the actual local entity export operation.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   export. For supported options see
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity().
   * @param array $context
   *   An associative array containing context related to the circumstances of
   *   the operation. See
   *   \Drupal\entity_sync\Export\ManagerInterface::exportLocalEntity().
   */
  protected function doExportLocalEntity(
    SyncInterface $sync,
    ContentEntityInterface $local_entity,
    array $options,
    array $context
  ) {
    // Initiate the operation.
    $this->initiate(
      Events::LOCAL_ENTITY_INITIATE,
      'export_entity',
      $context,
      $sync
    );

    // Export the entity.
    $data = $this->createOrUpdate($local_entity, $sync, $options);

    // Terminate the operation.
    // Add to the context the local entity that was exported.
    $this->terminate(
      Events::LOCAL_ENTITY_TERMINATE,
      'export_entity',
      $context,
      $sync,
      $data ?? []
    );
  }

  /**
   * Exports the entity without halting execution if an exception is thrown.
   *
   * An error is logged instead; the caller may then continue with exporting the
   * next entity, if there is one.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   import. For supported options see the operations that we may be running:
   *   - \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalListByEntities()
   *   - \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity()
   * @param string $operation
   *   The operation that is doing the import; used for logging purposes.
   */
  protected function tryWrapDoExportLocalEntity(
    ContentEntityInterface $local_entity,
    SyncInterface $sync,
    array $options,
    $operation
  ) {
    try {
      $this->wrapDoExportLocalEntity($sync, $local_entity, $options);
    }
    catch (\Exception $e) {
      $this->logger->error(
        sprintf(
          'An "%s" exception was thrown while exporting the local entity with ID "%s" as part of the "%s" synchronization and the "%s" operation. The error message was: %s',
          get_class($e),
          $local_entity->id(),
          $sync->id(),
          $operation,
          $e->getMessage()
        )
      );
    }
  }

  /**
   * Export the changes contained in the given local entity to a remote entity.
   *
   * If an associated remote entity is identified, the remote entity will be
   * updated. A new remote entity will be created otherwise.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   export. For supported options see
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity().
   */
  protected function createOrUpdate(
    ContentEntityInterface $local_entity,
    SyncInterface $sync,
    array $options
  ) {
    // Build the entity mapping for this local entity.
    $entity_mapping = $this->localEntityMapping($local_entity, $sync);

    // If the entity mapping is empty we will not be updating or creating a
    // remote entity; nothing to do. Still, event subscribers may require an
    // action so we set it to skip.
    if (!$entity_mapping) {
      return ['action' => EntityManagerInterface::ACTION_SKIP];
    }

    $data = ['action' => $entity_mapping['action']];

    // Skip exporting the local entity if we are explicitly told to do so.
    if ($entity_mapping['action'] === EntityManagerInterface::ACTION_SKIP) {
      return $data;
    }
    elseif ($entity_mapping['action'] === EntityManagerInterface::ACTION_CREATE) {
      $data['response'] = $this->create(
        $local_entity,
        $sync,
        $entity_mapping,
        $options
      );
      return $data;
    }
    elseif ($entity_mapping['action'] === EntityManagerInterface::ACTION_UPDATE) {
      $data['response'] = $this->update(
        $local_entity,
        $sync,
        $entity_mapping,
        $options
      );
      return $data;
    }
    else {
      throw new \RuntimeException(
        sprintf(
          'Unsupported entity mapping action "%s"',
          $entity_mapping['action']
        )
      );
    }
  }

  /**
   * Export the changes from the given local entity to a new remote entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param array $entity_mapping
   *   An associative array containing information about the remote entity being
   *   mapped to the given local entity.
   *   See \Drupal\entity_sync\Export\Event\LocalEntityMapping::entityMapping.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   export. For supported options see
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity().
   *
   * @return mixed|null
   *   The response from the client, or NULL if the operation was not run.
   */
  protected function create(
    ContentEntityInterface $local_entity,
    SyncInterface $sync,
    array $entity_mapping,
    array $options
  ) {
    $settings = $sync->getOperationsSettings();
    if (empty($settings['export_entity']['create_entities'])) {
      return;
    }

    // Prepare the fields to create on the remote object.
    // phpcs:disable
    // @I Pass context to the field manager
    //    type     : improvement
    //    priority : normal
    //    labels   : field, export
    // phpcs:enable
    $remote_fields = $this->fieldManager->export(
      $local_entity,
      NULL,
      $sync,
      $options
    );

    $client_options = $options['client'] ?? [];
    if (!$client_options) {
      $client_options = $sync->getRemoteResourceSettings()['client']['options'] ?? [];
    }

    // Do the export i.e. call the client.
    return $this->clientFactory
      ->getByClientConfig($entity_mapping['client'])
      ->create($remote_fields, $client_options);
  }

  /**
   * Export the changes from the given local entity to the remote entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   * @param array $entity_mapping
   *   An associative array containing information about the remote entity being
   *   mapped to the given local entity.
   *   See \Drupal\entity_sync\Export\Event\LocalEntityMapping::entityMapping.
   * @param array $options
   *   An associative array of options that determine various aspects of the
   *   export. For supported options see
   *   \Drupal\entity_sync\Export\EntityManagerInterface::exportLocalEntity().
   *
   * @return mixed|null
   *   The response from the client, or NULL if the operation was not run.
   * phpcs:disable
   * @I Support entity update validation
   *    type     : bug
   *    priority : normal
   *    labels   : export, validation
   * @I Check if the changes have already been exported
   *    type     : improvement
   *    priority : normal
   *    labels   : export, validation
   * phpcs:enable
   */
  protected function update(
    ContentEntityInterface $local_entity,
    SyncInterface $sync,
    array $entity_mapping,
    array $options
  ) {
    $settings = $sync->getOperationsSettings();
    if (empty($settings['export_entity']['update_entities'])) {
      return;
    }

    // Prepare the fields to update on the remote object.
    // phpcs:disable
    // @I Pass context to the field manager
    //    type     : improvement
    //    priority : normal
    //    labels   : field, export
    // phpcs:enable
    $remote_fields = $this->fieldManager->export(
      $local_entity,
      $entity_mapping['id'],
      $sync,
      $options
    );

    $client_options = $options['client'] ?? [];
    if (!$client_options) {
      $client_options = $sync->getRemoteResourceSettings()['client']['options'] ?? [];
    }

    // Do the export i.e. call the client.
    return $this->clientFactory
      ->getByClientConfig($entity_mapping['client'])
      ->update($entity_mapping['id'], $remote_fields, $client_options);
  }

  /**
   * Builds and returns the remote ID for the given local entity.
   *
   * The local entity mapping defines if and which remote entity this local
   * entity will be exported to. The default mapping identifies the remote
   * entity based on a local entity field containing the remote entity's ID.
   *
   * An event is dispatched that allows subscribers to map the local entity to a
   * different remote entity, or to decide to not export it at all.
   *
   * @param \Drupal\core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The configuration object for synchronization that defines the operation
   *   we are currently executing.
   *
   * @return array
   *   The final entity mapping.
   */
  protected function localEntityMapping(
    ContentEntityInterface $local_entity,
    SyncInterface $sync
  ) {
    $event = new LocalEntityMappingEvent(
      $local_entity,
      $sync
    );
    $this->eventDispatcher->dispatch($event, Events::LOCAL_ENTITY_MAPPING);

    // Return the final mapping.
    return $event->getEntityMapping();
  }

  /**
   * Sorts the synchronizations being queued.
   *
   * The synchronizations are sorted based on the `queue_weight` setting on the
   * `export_entity` operation. The weight defaults to 0 for synchronizations
   * that the setting is not set.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface[] $syncs
   *   The synchronizations being queued.
   *
   * @return \Drupal\entity_sync\Entity\SyncInterface[]
   *   The sorted synchronizations.
   */
  protected function sortQueueSynchronizations(array $syncs) {
    uasort(
      $syncs,
      function ($a, $b) {
        $a_weight = $a->getOperationsSettings()['export_entity']['queue_weight'] ?? 0;
        $b_weight = $b->getOperationsSettings()['export_entity']['queue_weight'] ?? 0;

        if ($a_weight === $b_weight) {
          return 0;
        }

        return ($a_weight < $b_weight) ? -1 : 1;
      }
    );

    return $syncs;
  }

  /**
   * Gives the opportunity to subscribers to respond to and/or cancel queueing.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The Synchronization or Operation Type that defines the operation that
   *   we are queueing.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being queued for export.
   * @param array $context
   *   An associative array containing the queueing context. See
   *   \Drupal\entity_sync\Export\EntityManagerInterface::queueExportLocalEntityAllSyncs().
   */
  protected function shouldCancelQueueing(
    SyncInterface $sync,
    ContentEntityInterface $entity,
    array $context
  ) {
    $operation = 'export_entity';
    $event = new QueueOperationEvent(
      $operation,
      $context + ['local_entity' => $entity],
      $sync
    );
    $this->eventDispatcher->dispatch($event, Events::LOCAL_ENTITY_QUEUE);

    [$cancel, $messages] = $event->getCancellations();
    if (!$cancel) {
      return FALSE;
    }

    foreach ($messages as $message) {
      $this->logger->warning(
        sprintf(
          'Queueing the "%s" operation for the synchronization with ID "%s" and for the entity with ID "%s" was cancelled with message: %s',
          $operation,
          $sync->id(),
          $entity->id(),
          $message
        )
      );
    }

    return TRUE;
  }

  /**
   * Creates as queue item for the given data.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The Synchronization or Operation Type entity that defines the operation
   *   that will be queued.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity that the operation relates to.
   * @param string $queue
   *   The queue to add the item to.
   */
  protected function createQueueItem(
    SyncInterface $sync,
    ContentEntityInterface $entity,
    string $queue
  ) {
    if ($sync->getEntityTypeId() === 'entity_sync_operation_type') {
      $this->queueFactory
        ->get($queue)
        ->createItem($this->buildOperationQueueItem($sync, $entity));
      return;
    }

    $this->queueFactory
      ->get($queue)
      ->createItem($this->buildSynchronizationQueueItem($sync, $entity));
  }

  /**
   * Creates a queue item for the given operation type and entity.
   *
   * @param \Drupal\entity_sync\Entity\OperationTypeInterface $operation_type
   *   The Operation Type entity that defines the operation that will be queued.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity that the operation relates to.
   *
   * @return array
   *   An associative array containing the data that the queue item requires for
   *   its processing.
   */
  protected function buildOperationQueueItem(
    OperationTypeInterface $operation_type,
    ContentEntityInterface $entity
  ) {
    $storage = $this->entityTypeManager->getStorage('entity_sync_operation');
    $operation = $storage->create([
      OperationField::TYPE => $operation_type->id(),
      OperationField::ENTITY => $entity,
    ]);

    if (!$operation->hasField(OperationField::ENTITY)) {
      throw new \RuntimeException(sprintf(
        'Operations of type "%s" do not have a field for storing the related entity.',
        $operation_type->id()
      ));
    }

    $storage->save($operation);

    return [
      'sync_id' => $operation_type->id(),
      'operation_id' => $operation->id(),
    ];
  }

  /**
   * Creates a queue item for the given synchronization and entity.
   *
   * @param \Drupal\entity_sync\Entity\SyncInterface $sync
   *   The Synchronization entity that defines the operation that will be
   *   queued.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity that the operation relates to.
   *
   * @return array
   *   An associative array containing the data that the queue item requires for
   *   its processing.
   */
  protected function buildSynchronizationQueueItem(
    SyncInterface $sync,
    ContentEntityInterface $entity
  ) {
    return [
      'sync_id' => $sync->id(),
      'entity_type_id' => $entity->getEntityTypeId(),
      'entity_id' => $entity->id(),
    ];
  }

}
